import os
import re
import glob
import ROOT
import math
import zfit
import numpy
import pprint
import extset
import utils_noroot      as utnr
import zutils.utils      as zut
import matplotlib.pyplot as plt

from zutils.plot   import plot     as zfp
from fitter        import zfitter
from log_store     import log_store

log = log_store.add_logger(name='rk_extractor:normalizer')
#--------------------------------------
class normalizer:
    '''
    Class used to find normalizations for combinatorial and PRec components
    in the high-q2 signal model
    '''
    #--------------------------------------
    def __init__(self, dset=None, trig=None, d_model=None, d_val=None, d_var=None):
        self._d_model = d_model 
        self._dset    = dset
        self._trig    = trig
        self._d_val   = d_val
        self._d_var   = d_var

        self._l_flt_par= ['lm_cb', 'mu_cb', 'a', 'b', 'ncb', 'npr_ee', 'nsg_ee']
        self._dat_path = f'{os.environ["CASDIR"]}/tools/apply_selection/blind_fits/data/v10.21p2'

        self._d_const  = {}
        self._nbins    = 64 
        self._rng_mm   = 5180, 5600 
        self._rng_ee   = 4500, 6100 
        self._bln_ee   = 5000, 5450
        self._bln_mm   = 5200, 5360

        self._d_data      = None
        self._s_par       = None
        self._d_pre       = None
        self._out_dir     = None
        self._l_bdt_bin   = None 
        self._custom_data = False
        self._initialized = False 
    #--------------------------------------
    def _initialize(self):
        if self._initialized:
            return

        os.makedirs(f'{self._out_dir}/pdf', exist_ok=True)

        self._l_bdt_bin = self._get_bdt_bins()
        self._s_par     = self._get_parameters() 
        self._d_pre     = {par.name : par.value().numpy() for par in self._s_par}
        self._d_const   = self._prepare_pars()
        self._d_data    = self._get_data()

        self._initialized = True 
    #--------------------------------------
    def _get_parameters(self):
        s_par = set()
        for pdf_mm, pdf_ee in self._d_model.values():
            s_mod_par_mm = pdf_mm.get_params()
            s_mod_par_ee = pdf_ee.get_params()
            s_par        = s_par.union(s_mod_par_mm)
            s_par        = s_par.union(s_mod_par_ee)

        return s_par
    #--------------------------------------
    @property
    def data(self):
        return self._d_data

    @data.setter
    def data(self, value):
        self._custom_data = True
        self._d_data      = value
    #--------------------------------------
    def _get_bdt_bins(self):
        l_mod_key  = [ key for key in self._d_model ]
        l_bdt_sbin = [ re.match('.*_(\d)$', mod_key).groups()[0] for mod_key in l_mod_key ]
        l_bdt_ibin = [ int(sbin) for sbin in l_bdt_sbin ]

        return l_bdt_ibin
    #--------------------------------------
    def _get_data_paths(self):
        if   self._dset == 'r1':
            l_dset = [f'{self._dat_path}/2011_{self._trig}', f'{self._dat_path}/2012_{self._trig}']
        elif self._dset == 'r2p1':
            l_dset = [f'{self._dat_path}/2015_{self._trig}', f'{self._dat_path}/2016_{self._trig}']
        elif self._dset == 'all':
            l_dset = [f'{self._dat_path}/{year}_{self._trig}' for year in [2011, 2012, 2015, 2016, 2017, 2018] ]
        else:
            l_dset = [f'{self._dat_path}/{self._dset}_{self._trig}']

        l_wc = [f'{dset}/*.root' for dset in l_dset]

        l_l_path = [ glob.glob(wc) for wc in l_wc ]
        for l_path, wc in zip(l_l_path, l_wc):
            if len(l_path) == 0:
                log.error(f'No file found in {wc}')
                raise

        return l_wc
    #-------------------------------------
    def _filter(self, rdf, ibin):
        d_bdt_wp, _, _ = extset.get_bdt_bin_settings(ibin)

        [lo_cmb, hi_cmb] = d_bdt_wp['BDT_cmb']
        [lo_prc, hi_prc] = d_bdt_wp['BDT_prc']

        hi_cmb = +10 if hi_cmb is None else hi_cmb
        hi_prc = +10 if hi_prc is None else hi_prc

        lo_cmb = -10 if lo_cmb is None else lo_cmb
        lo_prc = -10 if lo_prc is None else lo_prc

        rdf = rdf.Filter(f'(BDT_cmb > {lo_cmb}) && (BDT_cmb < {hi_cmb})', 'CMB')
        rdf = rdf.Filter(f'(BDT_prc > {lo_prc}) && (BDT_prc < {hi_prc})', 'PRC')

        rep = rdf.Report()
        rep.Print()

        return rdf
    #-------------------------------------
    def _get_mass(self, l_dpath, ibin):
        rdf = ROOT.RDataFrame(self._trig, l_dpath)
        rdf = self._filter(rdf, ibin)

        arr_mass = rdf.AsNumpy(['B_M'])['B_M']

        return arr_mass.tolist()
    #--------------------------------------
    @utnr.timeit
    def _get_data(self):
        dat_path = f'{self._out_dir}/data/{self._dset}_{self._trig}.json'
        if   self._d_data is not None:
            log.warning(f'Using data passed by user')
            return self._d_data
        elif os.path.isfile(dat_path):
            log.info(f'Loading cached data from: {dat_path}')
            d_dat = utnr.load_json(dat_path)
            return d_dat 

        log.info(f'Caching data: {dat_path}')

        os.makedirs(f'{self._out_dir}/data', exist_ok=True)

        l_dpath = self._get_data_paths()
        d_dat   = { f'bdt_{ibin}' : self._get_mass(l_dpath, ibin) for ibin in self._l_bdt_bin}

        log.info(f'Saving to: {dat_path}')
        utnr.dump_json(d_dat, dat_path) 

        return d_dat 
    #--------------------------------------
    def _plot_data(self, arr_mass):
        if not self._out_dir:
            return

        rng_ee = 4800, 6000
        rng_mm = 5100, 5600
        rng = rng_mm if self._trig == 'MTOS' else rng_ee
        plt.hist(arr_mass, range=rng, bins=50, histtype='step')
        plot_path = f'{self._out_dir}/data/{self._dset}_{self._trig}.png'
        log.info(f'Saving to: {plot_path}')
        plt.savefig(plot_path)
        plt.close('all')
    #--------------------------------------
    def _prepare_pars(self):
        for par in self._s_par:
            par.floating = False

        d_const = {}
        for par in self._s_par:
            for flt_par in self._l_flt_par:
                if par.name.startswith(flt_par):
                    par.floating = True
                else:
                    continue

                if par.name in self._d_val:
                    var = self._d_var[par.name]
                    val = self._d_val[par.name]
                    d_const[par.name] = val, math.sqrt(var)

        return d_const
    #--------------------------------------
    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    def out_dir(self, value):
        try:
            os.makedirs(value, exist_ok=True)
        except:
            log.error(f'Cannot create directory: {value}')
            raise

        self._out_dir = value
    #--------------------------------------
    def _get_stats(self, arr_dat):
        if self._trig != 'MTOS':
            return ''

        arr_flg = (arr_dat > 5180) & (arr_dat < 5400)
        arr_dat = arr_dat[arr_flg]
        ndata   = float(arr_dat.size)

        s_par     = self._model.get_params(floating=False)
        [pst_val] = [ par.value().numpy() for par      in s_par               if par.name.startswith('nsg_mm_') ]
        [pre_val] = [ val                 for nam, val in self._d_pre.items() if nam.startswith('nsg_mm_')      ]

        v1 = f'Data: {ndata:.0f} $m\in [5180, 5400]$'
        v2 = f'Fitted: {pst_val:.0f}'
        v3 = f'Expected: {pre_val:.0f}'

        return f'{v1}\n{v2}\n{v3}'
    #--------------------------------------
    def _get_pdf_names(self):
        d_leg         = {}
        d_leg['prc']  = r'$\psi(2S) + c\bar{c}$'
        d_leg['bpks'] = r'$B^+\to K^{*+}e^+e^-$'
        d_leg['bdks'] = r'$B^0\to K^{*0}e^+e^-$'
        d_leg['bsph'] = r'$B_s\to \phi e^+e^-$'
        d_leg['bpk1'] = r'$B^+\to K_{1}e^+e^-$'
        d_leg['bpk2'] = r'$B^+\to K_{2}e^+e^-$'

        return d_leg
    #--------------------------------------
    def _get_blinding(self):
        if self._trig == 'MTOS':
            return

        low, hig= self._bln_ee
        l_blnd  = [extset.sig_name, low, hig]

        return l_blnd
    #--------------------------------------
    def _plot_fit(self, data=None, model=None, name=None, result=None, stacked=None):
        stats   = self._get_stats(data)
        rng     = self._rng_mm if self._trig == 'MTOS' else self._rng_ee
        l_blnd  = self._get_blinding()

        obj= zfp(data=data, model=model, result=None)
        obj.plot(skip_pulls=False, blind=l_blnd, nbins=self._nbins, ranges=[rng], d_leg=self._get_pdf_names(), stacked=stacked, ext_text=stats)
        obj.axs[0].grid()

        if not self._custom_data:
            if   '_all_' in name and 'MTOS' in name:
                obj.axs[0].set_ylim(0,  250)
            elif '_all_' in name and 'ETOS' in name:
                obj.axs[0].set_ylim(0,  120)
            elif '_all_' in name and 'GTIS' in name:
                obj.axs[0].set_ylim(0,   50)

        obj.axs[1].set_ylim(-5, 5)
        obj.axs[1].axhline(0, color='r')

        os.makedirs(f'{self._out_dir}/fits', exist_ok=True)
        plot_path = f'{self._out_dir}/fits/{name}.png'
        log.info(f'Saving to: {plot_path}')
        plt.savefig(plot_path)
        plt.close('all')
    #--------------------------------------
    @utnr.timeit
    def _fit(self, d_mod, d_dat):
        tot_nll = None
        d_fdat  = {}
        d_fmod  = {}
        for bdt_bin in self._l_bdt_bin:
            _, mod = d_mod[f'bdt_{bdt_bin}']
            l_dat  = d_dat[f'bdt_{bdt_bin}']
            arr_mas= numpy.array(l_dat)
            obs    = mod.space
            dat    = zfit.Data.from_numpy(obs=obs, array=arr_mas)

            d_fdat[f'bdt_{bdt_bin}'] = dat
            d_fmod[f'bdt_{bdt_bin}'] = mod

            rng    = self._rng_mm if self._trig == 'MTOS' else self._rng_ee
            log.info(f'Adding nll for bin {bdt_bin} in range {rng}')
            nll    = zfit.loss.ExtendedUnbinnedNLL(model=mod, data=dat, fit_range=rng)

            tot_nll = nll if tot_nll is None else nll + tot_nll

        minimizer = zfit.minimize.Minuit()
        log.debug('Minimizing')
        result    = minimizer.minimize(tot_nll)

        return result, d_fdat, d_fmod
    #--------------------------------------
    def _print_pdfs(self, prefix):
        for key, (mod_mm, mod_ee) in self._d_model.items():
            zut.print_pdf(mod_mm, txt_path=f'{self._out_dir}/pdf/{prefix}_mm_{key}_{self._dset}_{self._trig}.txt', d_const=self._d_const)
            zut.print_pdf(mod_ee, txt_path=f'{self._out_dir}/pdf/{prefix}_ee_{key}_{self._dset}_{self._trig}.txt', d_const=self._d_const)
    #--------------------------------------
    def _plot_fits(self, d_fdat, d_fmod):
        for ibin in self._l_bdt_bin:
            key = f'bdt_{ibin}'
            arr_mass = d_fdat[key]
            model    = d_fmod[key]

            self._plot_fit(data=arr_mass, model=model, name=f'{self._trig}_{self._dset}_{key}_stk', stacked= True)
    #--------------------------------------
    @utnr.timeit
    def get_fit_result(self):
        self._initialize()

        self._print_pdfs('pre')
        res, d_fdat, d_fmod = self._fit(self._d_model, self._d_data)
        self._print_pdfs('pos')

        self._plot_fits(d_fdat, d_fmod)

        res.hesse()
        res.freeze()

        return res
#--------------------------------------




from importlib.resources import files
from rk.calc_utility     import getGeomEff          as geo_eff
from datetime            import datetime
from log_store           import log_store

import os
import re
import math
import glob
import ROOT
import numpy
import utils
import mplhep
import pprint
import pandas            as pnd
import jacobi            as jac
import utils_noroot      as utnr
import matplotlib.pyplot as plt

log=log_store.add_logger(name='rk_extractor:scales')
#------------------------------------------
def get_proc_labels():
    d_proc_lab         = dict() 
    d_proc_lab['bpkp'] = r'$B^+\to K^+ e^+e^-$'
    d_proc_lab['bpks'] = r'$B^+\to K^{*+}(\to K\pi^0) e^+e^-$'
    d_proc_lab['bdks'] = r'$B^+\to K^{*0}(\to K\pi) e^+e^-$'
    d_proc_lab['bsph'] = r'$B_s\to \phi(\to K K) e^+e^-$'
    d_proc_lab['bpk1'] = r'$B^+\to K_1(\to K \pi\pi) e^+e^-$'
    d_proc_lab['bpk2'] = r'$B^+\to K_2(\to K \pi\pi) e^+e^-$'

    return d_proc_lab
#------------------------------------------
class eff_calc:
    def __init__(self, proc=None, year=None, trig=None):
        self._l_proc = ['bpkp', 'bpks', 'bdks', 'bsph', 'bpk1', 'bpk2']  if proc is None else proc
        self._l_trig = ['ETOS', 'GTIS']                                  if trig is None else trig
        self._l_year = ['2011', '2012', '2015', '2016', '2017', '2018']  if year is None else year

        self._d_proc_lab = get_proc_labels()
        self._d_geo_eff  = {}
        self._d_sel_tis  = {} 
        self._d_sel_tos  = {} 
        self._dvers      = 'v10.21p2'
        self._min_date   = '01.09.23'
        self._out_dir    = None

        plt.style.use(mplhep.style.LHCb2)

        self._initialized=False
    #------------------------------------------
    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    def out_dir(self, value):
        try:
            os.makedirs(value, exist_ok=True)
        except:
            log.error(f'Cannot make: {value}')
            raise

        self._out_dir = value
    #------------------------------------------
    def get_efficiencies(self, d_sel=None):
        '''
        Parameters
        ---------------------
        d_sel (dict): If set, it will read the original files and apply the selection, e.g. {'BDT_cmb' : 'BDT_cmb > 0.977'}
        '''
        d_eff = {}
        log.info('-' * 20)
        log.info(f'{"Process":<15}{"Trigger":<15}{"Year":<15}')
        log.info('-' * 20)
        for proc in self._l_proc: 
            d_eff[proc] = {}
            for trig in self._l_trig: 
                d_eff[proc][trig] = {}
                for year in self._l_year: 
                    log.info(f'{proc:<15}{trig:<15}{year:<15}')
                    d_eff[proc][trig][year] = self._get_efficiency(proc=proc, trig=trig, year=year, d_sel=d_sel)
        log.info('-' * 20)

        self._plot_geo_eff()
        self._plot_sel_eff()

        return d_eff
    #------------------------------------------
    def _get_df_from_dict(self, d_data, ext_name=''):
        df = pnd.DataFrame(columns=['Process', 'Year', 'Value', 'Error'])
        for proc in self._l_proc:
            for year in self._l_year:
                key      = f'{proc}{ext_name}_{year}'
                val, err = d_data[key] 

                df = utnr.add_row_to_df(df, [proc, year, val, err] )

        return df
    #------------------------------------------
    def _plot_geo_eff(self):
        if self._out_dir is None:
            return

        df = self._get_df_from_dict(self._d_geo_eff, ext_name='_ee')

        plt_dir = f'{self._out_dir}/plots'
        os.makedirs(plt_dir, exist_ok=True)

        ax=None
        for proc, df_p in df.groupby('Process'):
            ax=df_p.plot(x='Year', y='Value', yerr='Error', ax=ax, label=self._d_proc_lab[proc], marker='o')

        plt_path = f'{plt_dir}/geo_eff.png'
        plt.grid()
        plt.ylim(0.00, 0.40)
        plt.savefig(plt_path)
        plt.close('all')
    #------------------------------------------
    def _plot_sel_eff(self):
        if self._out_dir is None:
            return

        plt_dir = f'{self._out_dir}/plots'
        os.makedirs(plt_dir, exist_ok=True)
        for trig, d_data in {'GTIS' : self._d_sel_tis , 'ETOS' : self._d_sel_tos}.items():
            if trig not in self._l_trig:
                continue
            df = self._get_df_from_dict(d_data)

            ax=None
            for proc, df_p in df.groupby('Process'):
                ax=df_p.plot(x='Year', y='Value', yerr='Error', ax=ax, label=self._d_proc_lab[proc], marker='o')

            plt_path = f'{plt_dir}/sel_eff_{trig}.png'
            plt.grid()
            plt.savefig(plt_path)
            plt.close('all')
    #------------------------------------------
    def _switch_sample(self, proc, year):
        '''
        Switch samples for unavailable sim08 generator efficiencies
        '''
        if  proc in ['bpk1_ee', 'bpk2_ee'] and year in ['2011', '2012', '2015', '2016']:
            new_year = '2017'
            log.warning(f'Gen efficiency using {year} -> {new_year} for {proc}')
            year = new_year 

        return proc, year
    #------------------------------------------
    def _get_geo_eff(self, proc, year):
        key        = f'{proc}_{year}'
        proc, year = self._switch_sample(proc, year)

        if key in self._d_geo_eff:
            return self._d_geo_eff[key]
            
        self._d_geo_eff[key] = geo_eff(proc, year)

        return self._d_geo_eff[key]
    #------------------------------------------
    def _get_efficiency(self, proc=None, trig=None, year=None, d_sel=None):
        sel_eff  = self._get_rec_eff(d_sel,  proc, trig, year)
        geo_acc  = self._get_geo_eff(f'{proc}_ee',       year)

        val, var = jac.propagate(lambda x : x[0] * x[1], [geo_acc[0], sel_eff[0]], [[geo_acc[1] ** 2, 0], [0, sel_eff[1] ** 2]]) 
        val      = float(val)
        err      = math.sqrt(var)

        if trig == 'GTIS':
            self._d_sel_tis[f'{proc}_{year}'] = val, err
        else:
            self._d_sel_tos[f'{proc}_{year}'] = val, err

        return val, err
    #------------------------------------------
    def _get_sel_yield(self, proc, trig, year, d_sel):
        cas_dir  = os.environ['CASDIR']
        root_wc  = f'{cas_dir}/tools/apply_selection/rare_backgrounds/{proc}/{self._dvers}/{year}_{trig}/*.root'
        rdf      = ROOT.RDataFrame(trig, root_wc)

        if d_sel is None:
            return rdf.Count().GetValue()

        for key, cut in d_sel.items():
            rdf = rdf.Filter(cut, key)

        rep=rdf.Report()
        rep.Print()

        return rdf.Count().GetValue()
    #------------------------------------------
    def _get_date_flag(self, date):
        l_date = re.findall('\d{2}\.\d{2}\.\d{2}', date)
        if len(l_date) == 0:
            return False

        dstr = datetime.strptime

        return any( dstr(self._min_date, '%d.%m.%y') < dstr(date, '%d.%m.%y') for date in l_date)
    #------------------------------------------
    def _switch_yields(self, proc, year):
        if   proc in ['bpk1', 'bpk2'] and year in ['2015', '2016']:
            new_year = '2017'
            log.warning(f'Reco yields using {year} -> {new_year} for {proc}')
            year = new_year
        elif proc == 'bpk2' and year == '2011':
            new_year = '2012'
            log.warning(f'Reco yields using {year} -> {new_year} for {proc}')
            year = new_year

        return proc, year
    #------------------------------------------
    def _get_year_entries(self, df, proc, year):
        proc, year = self._switch_yields(proc, year)

        df_y = df[df.Year == int(year)]
        if len(df_y) not in [1, 2]:
            log.error(f'Found more than two or fewer than one entries (polarities) after fully filtering')
            log.info(df)
            log.info('--->')
            log.info(df_y)
            raise

        l_pol=df_y.Polarity.tolist()
        s_pol=set(l_pol)
        if   len(df_y) == 2 and s_pol != {'MagUp', 'MagDown'}:
            log.error(f'Wrong polarities: {s_pol}')
            raise ValueError
        elif len(df_y) == 1:
            log.warning(f'Found only polarity: {s_pol}')

        return df_y.Events.sum()
    #------------------------------------------
    def _get_rec_yield(self, proc, year):
        ganga_dir = os.environ['GANDIR']
        json_path = f'{ganga_dir}/job_stats/{proc}.json'

        df = pnd.read_json(json_path)
        df = df.loc[[ self._get_date_flag(date) for date in df.Dates ]]

        yld= self._get_year_entries(df, proc, year)

        return yld 
    #------------------------------------------
    def _get_rec_eff(self, d_sel, proc, trig, year):
        sel = self._get_sel_yield(proc, trig, year, d_sel)
        rec = self._get_rec_yield(proc,       year)

        eff, eup, edn = utils.get_eff_err(sel, rec)

        err = (eup + edn) / 2.

        return eff, err
#------------------------------------------
class scales:
    def __init__(self, dset=None, trig=None, kind=None, vers=None, d_sel=None):
        self._dset = dset
        self._trig = trig 
        self._kind = kind 
        self._vers = vers
        self._d_sel= d_sel

        self._l_kind = ['bpks', 'bdks', 'bsph', 'bpk1', 'bpk2']
        self._l_trig = ['ETOS', 'GTIS']
        self._l_dset = ['2011', '2012', '2015', '2016', '2017', '2018']

        self._d_frbf = None
        self._d_eff  = None

        self._initialized = False
    #------------------------------------------
    def _check_arg(self, l_val, val, name):
        if val not in l_val:
            log.error(f'{name} {val} not allowed')
            raise ValueError
        else:
            log.debug(f'{name:<20}{"->":20}{val:<20}')
    #------------------------------------------
    def _initialize(self):
        if self._initialized:
            return

        log.debug('Initializing')

        if self._vers is None:
            log.error('Version not specified')
            raise

        log.debug(f'Using rare-signal scales version: {self._vers}')
        frbf_path    = files('extractor_data').joinpath(f'rare_sf/{self._vers}/fr_bf.json')
        self._d_frbf = utnr.load_json(frbf_path)

        self._d_eff  = self._get_efficiencies() 

        self._check_arg(self._l_kind, self._kind, 'Background')
        self._check_arg(self._l_dset, self._dset, 'Dataset')
        self._check_arg(self._l_trig, self._trig, 'Trigger')

        self._initialized = True
    #------------------------------------------
    def _get_efficiencies(self):
        log.debug('Getting efficiencies')

        eff_path     = files('extractor_data').joinpath(f'rare_sf/{self._vers}/eff_real.json'  )
        if not os.path.isfile(eff_path):
            log.debug('Not found efficiencies file, remaking it')
            calc   = eff_calc()
            d_eff  = calc.get_efficiencies(d_sel=self._d_sel)
            utnr.dump_json(d_eff, eff_path)
            log.debug(f'Efficiencies file saved to: {eff_path}')
        else:
            d_eff  = utnr.load_json(eff_path)
            log.debug('Found efficiencies file, not remaking it')

        return d_eff
    #------------------------------------------
    def _get_fr(self, kind):
        key= 'bpkp' if kind == 'sig' else self._kind
        fx = {'bpkp' : 'fu', 'bpks' : 'fu', 'bdks' : 'fd', 'bsph' : 'fs', 'bpk1' : 'fu', 'bpk2' : 'fu'}[key]
        fx = self._d_frbf['fr'][fx]

        return fx
    #------------------------------------------
    def _mult_brs(self, br_1, br_2):
        l_br_val = [br_1[0], br_2[0]]
        br_cov   = [[br_1[1] ** 2, 0], [0, br_2[1] ** 2]]
        val, var = jac.propagate(lambda x : x[0] * x[1], l_br_val, br_cov)

        return val, math.sqrt(var)
    #------------------------------------------
    def _get_br(self, kind):
        key= 'bpkp' if kind == 'sig' else self._kind

        if   key in ['bpkp', 'bpk1', 'bpk2']:
            br = self._d_frbf['bf'][key]
        elif key == 'bpks':
            br = self._mult_brs(self._d_frbf['bf'][key], self._d_frbf['bf']['k+kp'])
        elif key == 'bdks':
            br = self._mult_brs(self._d_frbf['bf'][key], self._d_frbf['bf']['kokp'])
        elif key == 'bsph':
            br = self._mult_brs(self._d_frbf['bf'][key], self._d_frbf['bf']['phkk'])
        else:
            log.error(f'Invalid key: {key}')
            raise

        return br
    #------------------------------------------
    def _get_ef(self, kind):
        key = 'bpkp' if kind == 'sig' else self._kind

        return self._d_eff[key][self._trig][self._dset]
    #------------------------------------------
    def _print_vars(self, l_tup):
        log.debug('-' * 20)
        log.debug(f'{"Var":<20}{"Value":<20}{"Error":<20}')
        log.debug('-' * 20)
        for (val, err), name in zip(l_tup, ['fr sig', 'br sig', 'eff sig', 'fr bkg', 'br bkg', 'eff bkg']):
            log.debug(f'{name:<20}{val:<20.3e}{err:<20.3e}')
        log.debug('-' * 20)
    #------------------------------------------
    def get_scale(self):
        self._initialize()

        fr_sig = self._get_fr('sig') 
        br_sig = self._get_br('sig') 
        ef_sig = self._get_ef('sig') 

        fr_bkg = self._get_fr('bkg') 
        br_bkg = self._get_br('bkg') 
        ef_bkg = self._get_ef('bkg') 

        l_tup = [fr_sig, br_sig, ef_sig, fr_bkg, br_bkg, ef_bkg]
        l_val = [ tup[0] for tup in l_tup]
        l_err = [ tup[1] for tup in l_tup]
        cov   = numpy.diag(l_err) ** 2

        self._print_vars(l_tup)

        val, var = jac.propagate(lambda x : (x[3] * x[4] * x[5]) / (x[0] * x[1] * x[2]), l_val, cov) 
        val = float(val)
        err = math.sqrt(var)

        return val, err 
#------------------------------------------

